"""Current version of tootgroup.py - Kept in a single place
for easier modification before every new release."""

CURRENT = "1.4.1"


def print_version():
    """Do just that"""
    print("tootgroup.py version: " + CURRENT)
